void mqttLoop() {
  if (!mqttClient.connected()) {
    connectAWSIoT();
  }
  mqttClient.loop();

}

void sendRequest(const char* request, const char* addlData) {

  stopAnimation();

  const size_t capacity = JSON_OBJECT_SIZE(4) + 200;
  DynamicJsonDocument doc(capacity);

  doc["deviceID"] = uniqId;
  doc["userID"] = userID;
  doc["request"] = request;
  doc["addlData"] = addlData;

  char buffer[512];
  size_t n = serializeJson(doc, buffer);
  //client.publish(PUBLISH_TOPIC, buffer, n);
  mqttClient.publish(PUBLISH_TOPIC, buffer);
  startAnimation();
}


void mqttCallback (char* topic, byte* payload, unsigned int length) {

  stopAnimation();

  Serial.print("Received. topic=");
  Serial.println(topic);
  Serial.print("Length=");
  Serial.println(length);

  if (strstr(topic, "timeInfo")) {
    handleTimeInfo(payload, length);
  }

  if (strstr(topic, "hourlyWeatherInfo")) {
    handleHourlyWeatherInfo(payload, length);
  }

  if (strstr(topic, "dailyWeatherInfo")) {
    handleDailyWeatherInfo(payload, length);
  }

  if (strstr(topic, "deviceCode")) {
    handleDeviceCodeInfo(payload, length);
  }

  if (strstr(topic, "deviceAuth")) {
    handleDeviceAuthInfo(payload, length);
  }

  if (strstr(topic, "schedule")) {
    handleScheduleInfo(payload, length);
  }

  if (strstr(topic, "news")) {
    handleNewsInfo(payload, length);
  }

  if (strstr(topic, "fitness")) {
    handleFitnessInfo(payload, length);
  }

  startAnimation();


}


void handleTimeInfo(byte* payload, unsigned int length) {

  Serial.println("Received Time Info");

  const size_t capacity = JSON_OBJECT_SIZE(3) + 80;
  DynamicJsonDocument doc(capacity);
  deserializeJson(doc, payload);

  daylightOffset_sec = doc["dstOffset"];
  gmtOffset_sec = doc["rawOffset"];



  syncReady = true;

  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);

  //once we have the time, we can detach and reset to sync every hour due to instability of RTC;
  tickerSyncTime.detach();
  int timeResync = 60 * 60;
  tickerSyncTime.attach(timeResync, requestTimeInfo);

  displayTime();
}

char hourlyWeatherHr[6][5];
char hourlyWeatherIcon[6][10];
int hourlyWeatherPrecip[6];
int hourlyWeatherTemp[6];

void handleHourlyWeatherInfo(byte* payload, unsigned int length) {

  Serial.println("Received Hourly Weather Info");

  int values = 7;

  const size_t capacity = JSON_ARRAY_SIZE(values) + values * JSON_OBJECT_SIZE(5) + 250;
  DynamicJsonDocument doc(capacity);
  DeserializationError err = deserializeJson(doc, payload);;

  if (err) {
    Serial.print(F("deserializeJson() failed: "));
    Serial.println(err.c_str());
    return;
  }

  JsonObject root_0 = doc[0];
  int temp = root_0["temp"]; // 71
  const char* icon = root_0["icon"]; // "clear"

  strcpy(hourlyWeatherHr[0], root_0["hr"]);
  strcpy(hourlyWeatherIcon[0], root_0["icon"]);
  hourlyWeatherTemp[0] = root_0["temp"];
  hourlyWeatherPrecip[0] = root_0["pp"];

  JsonObject root_1 = doc[1];
  strcpy(hourlyWeatherHr[1], root_1["hr"]);
  strcpy(hourlyWeatherIcon[1], root_1["icon"]);
  hourlyWeatherTemp[1] = root_1["temp"];
  hourlyWeatherPrecip[1] = root_1["pp"];

  JsonObject root_2 = doc[2];
  strcpy(hourlyWeatherHr[2], root_2["hr"]);
  strcpy(hourlyWeatherIcon[2], root_2["icon"]);
  hourlyWeatherTemp[2] = root_2["temp"];
  hourlyWeatherPrecip[2] = root_2["pp"];

  JsonObject root_3 = doc[3];
  strcpy(hourlyWeatherHr[3], root_3["hr"]);
  strcpy(hourlyWeatherIcon[3], root_3["icon"]);
  hourlyWeatherTemp[3] = root_3["temp"];
  hourlyWeatherPrecip[3] = root_3["pp"];

  JsonObject root_4 = doc[4];
  strcpy(hourlyWeatherHr[4], root_4["hr"]);
  strcpy(hourlyWeatherIcon[4], root_4["icon"]);
  hourlyWeatherTemp[4] = root_4["temp"];
  hourlyWeatherPrecip[4] = root_4["pp"];

  JsonObject root_5 = doc[5];
  strcpy(hourlyWeatherHr[5], root_5["hr"]);
  strcpy(hourlyWeatherIcon[5], root_5["icon"]);
  hourlyWeatherTemp[5] = root_5["temp"];
  hourlyWeatherPrecip[5] = root_5["pp"];

  displayCurrentWeather(icon, temp);

}

char dailyWeatherName[4][40];
char dailyWeatherIcon[4][12];
int dailyWeatherTemp[4];

void handleDailyWeatherInfo(byte* payload, unsigned int length) {

  Serial.println("Received Hourly Weather Info");

  int values = 4;

  const size_t capacity = JSON_ARRAY_SIZE(values) + values * JSON_OBJECT_SIZE(4) + 170;
  DynamicJsonDocument doc(capacity);

  deserializeJson(doc, payload);

  JsonObject root_0 = doc[0];
  strcpy(dailyWeatherName[0], root_0["name"]);
  strcpy(dailyWeatherIcon[0], root_0["icon"]);
  dailyWeatherTemp[0] = root_0["temp"];

  JsonObject root_1 = doc[1];
  strcpy(dailyWeatherName[1], root_1["name"]);
  strcpy(dailyWeatherIcon[1], root_1["icon"]);
  dailyWeatherTemp[1] = root_1["temp"];

  JsonObject root_2 = doc[2];
  strcpy(dailyWeatherName[2], root_2["name"]);
  strcpy(dailyWeatherIcon[2], root_2["icon"]);
  dailyWeatherTemp[2] = root_2["temp"];

  JsonObject root_3 = doc[3];
  strcpy(dailyWeatherName[3], root_3["name"]);
  strcpy(dailyWeatherIcon[3], root_3["icon"]);
  dailyWeatherTemp[3] = root_3["temp"];

}

void handleDeviceCodeInfo(byte* payload, unsigned int length) {

  Serial.println("Received MS Code Info");

  const size_t capacity = JSON_OBJECT_SIZE(2) + 100;
  DynamicJsonDocument doc(capacity);

  DeserializationError err = deserializeJson(doc, payload);;

  const char* code = doc["code"]; // "AG5XPZR2D"
  const char* url = doc["url"]; // "https://microsoft.com/devicelogin"
  //const char* message = doc["message"];

  Serial.println(url);
  Serial.println(code);

  displayDeviceCode(url, code);

}

void handleDeviceAuthInfo(byte* payload, unsigned int length) {

  Serial.println("Received Authenication Info");

  const size_t capacity = JSON_OBJECT_SIZE(2) + 70;
  DynamicJsonDocument doc(capacity);
  deserializeJson(doc, payload);

  bool success = doc["success"]; // true
  const char* message = doc["message"]; // "Authorization Successful!"

  if (success) {
    File scheduleAUTH = SPIFFS.open("/msAuth", FILE_WRITE);
    scheduleAUTH.print("authenticated");
    scheduleAUTH.close();
  } else {
    SPIFFS.remove("/msAuth");
  }

  displayDeviceAuth(message);

}

void handleScheduleInfo(byte* payload, unsigned int length) {


  Serial.println("Received Schedule Info");

  SPIFFS.remove("/schedule.json");
  delay(100);

  File scheduleJSON = SPIFFS.open("/schedule.json", FILE_WRITE);
  if (!scheduleJSON) {
    Serial.println("There was an error opening the file for writing");
    return;
  }

  for (int i = 0; i < length; i++) {
    scheduleJSON.print((char)payload[i]);
  }
  scheduleJSON.close();
  loadScheduleJSON(false);

}

void handleNewsInfo(byte* payload, unsigned int length) {

  Serial.println("Received News Info");

  SPIFFS.remove("/news.json");
  delay(100);

  File newsJSON = SPIFFS.open("/news.json", FILE_WRITE);
  if (!newsJSON) {
    Serial.println("There was an error opening the file for writing");
    return;
  }

  for (int i = 0; i < length; i++) {
    newsJSON.print((char)payload[i]);
  }
  newsJSON.close();

}

void handleFitnessInfo(byte* payload, unsigned int length) {

  Serial.println("Received Fitness Info");

  SPIFFS.remove("/fitness.json");
  delay(100);

  File fitnessJSON = SPIFFS.open("/fitness.json", FILE_WRITE);
  if (!fitnessJSON) {
    Serial.println("There was an error opening the file for writing");
    return;
  }

  for (int i = 0; i < length; i++) {
    fitnessJSON.print((char)payload[i]);
  }
  fitnessJSON.close();

}

/*
  void messageHandler(String &topic, String &payload) {
  Serial.println("incoming: " + topic + " - " + payload);


  if (topic.indexOf("timeInfo") > - 1 ) {

    Serial.println("Here!");
    const size_t capacity = JSON_OBJECT_SIZE(2) + 30;
    DynamicJsonDocument doc(capacity);
    deserializeJson(doc, payload);

    daylightOffset_sec = doc["dstOffset"];
    gmtOffset_sec = doc["rawOffset"];
    syncReady = true;

    configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);
    displayTime();

  }

    if (topic.indexOf("hourlyRequestInfo") > - 1 ) {

    Serial.println("Here!");


  }

  //  StaticJsonDocument<200> doc;
  //  deserializeJson(doc, payload);
  //  const char* message = doc["message"];
  }*/
