
int dayItem = 0;
char alarmDateTime[9] = "00000000";

void loadScheduleJSON(boolean showSchedule) {

  if (!isAuth(0)) {
    requestDeviceCode("ms");
    return;
  }

  File scheduleJSON = SPIFFS.open("/schedule.json", FILE_READ);

  if (!scheduleJSON) {
    Serial.println("The file does not exist");
    return;
  } else {
    size_t capacity = scheduleJSON.size() * 2;
    if ( capacity == 0 ) {
      Serial.println("The file is empty");
    } else {
      Serial.println("Getting data");
      DynamicJsonDocument doc(capacity);


      DeserializationError err = deserializeJson(doc, scheduleJSON);

      if (err) {
        Serial.print(F("deserializeJson() failed: "));
        Serial.println(err.c_str());
        return;
      }

      //serializeJson(doc, Serial);

      if (showSchedule) {

        allButtonsOff();

        char *scheduleArr[] = {"SCHEDULE"};
        lcarsBackground(st_LTPURPLE, scheduleArr, 0 , 0);

        int x = 10;
        int y = 40;
        int yTab = 50;
        uint16_t rowColor;

        tft.setTextDatum(TL_DATUM);
        int calendarEntryCount = 0;

        if (scheduleDay > 2) {
          scheduleDay = 0;
          dayItem = 0;
        }

        for (scheduleDay; scheduleDay < 3; scheduleDay++) {

          int dayEntryCount;
          JsonArray selectedDayArray;
          char dayEntryName[4];

          if (scheduleDay == 0) {

            dayEntryCount = doc["dayACount"];
            selectedDayArray = doc["dayA"];
            const char* dayAName = doc["dayAName"];
            strcpy(dayEntryName, dayAName);

          } else if (scheduleDay == 1) {

            dayEntryCount = doc["dayBCount"];
            selectedDayArray = doc["dayB"];
            const char* dayBName = doc["dayBName"];
            strcpy(dayEntryName, dayBName);

          } else if (scheduleDay == 2) {

            dayEntryCount = doc["dayCCount"];
            selectedDayArray = doc["dayC"];
            const char* dayCName = doc["dayCName"];
            strcpy(dayEntryName, dayCName);

          }

          int numOfEntries = 0;
          int origY = y;

          for (dayItem; dayItem < dayEntryCount; dayItem++) {
            tft.setTextDatum(TL_DATUM);

            if ( (calendarEntryCount % 2) == 0) {
              rowColor = st_WHITE;
            } else {
              rowColor = st_LTGRAY;
            }

            JsonObject dayObj = selectedDayArray[dayItem];

            const char* subject = dayObj["subject"];
            const char* startTime = dayObj["startTime"]; // "3:00 pm"
            const char* endTime = dayObj["endTime"]; // "3:30 pm"

            tft.setTextColor(rowColor, st_BLACK);
            tft.loadFont(STFONT24);
            tft.drawString(subject, x + 60, y, 4);

            tft.loadFont(STFONT24);
            tft.drawString(startTime, x + 60, y + 20, 4);
            tft.drawString(endTime, x + 140, y + 20, 4);

            y += yTab;

            calendarEntryCount++;
            numOfEntries++;

            if (calendarEntryCount == 5) {
              drawDayLine(numOfEntries, x, origY, dayEntryName);
              if (dayItem == dayEntryCount - 1) { //we got everything in this day and we need to force a move forward
                dayItem = 0;
                scheduleDay = scheduleDay + 1;
              } else { //increment the dayItem;
                dayItem++;
              }
              scheduleJSON.close();
              delay(20);
              return;
            }

          }
          //we got all the records for the day, and still have room for more
          drawDayLine(numOfEntries, x, origY, dayEntryName);
          dayItem = 0;
          numOfEntries = 0;
        }
      } //else {

      int alarmX = 10;
      int alarmY = 40;
      int alarmYTab = 50;
      boolean drewBackground = false;

      for (int i = 0; i < 3; i++) {

        int eventCount;
        JsonArray searchDayArray;
        int alarmDisplayCount = 0;


        if (i == 0) {
          searchDayArray = doc["dayA"];
          eventCount = doc["dayACount"];
        } else if (i == 1) {
          searchDayArray = doc["dayB"];
          eventCount = doc["dayBCount"];
        } else if (i == 2) {
          searchDayArray = doc["dayC"];
          eventCount = doc["dayCCount"];
        }

        for (int j = 0; j < eventCount; j++) {

          JsonObject possibleAlarmObj = searchDayArray[j];

          const char* alarmSubject = possibleAlarmObj["subject"];
          const char* alarmStartTime = possibleAlarmObj["startTime"]; // "3:00 pm"
          const char* alarmEndTime = possibleAlarmObj["endTime"]; // "3:30 pm"
          const char* startString = possibleAlarmObj["startString"];
          boolean reminder = possibleAlarmObj["reminder"];


          if (startString && reminder) {

            long startStringInt = atoi(startString);
            long alarmDateTimeInt = atoi(alarmDateTime);

            if (alarmDateTimeInt == startStringInt) {

               if (!drewBackground) {
                 allButtonsOff();

                 char *alarmArr[] = {"UP NEXT"};
                 lcarsBackground(st_LTPURPLE, alarmArr, 0 , 0);
                 drewBackground = true;
                 alarmBeep();
                }

                if (alarmDisplayCount < 5) {

                 tft.setTextColor(st_WHITE, st_BLACK);
                 tft.loadFont(STFONT24);
                 tft.drawString(alarmSubject, alarmX + 60, alarmY, 4);

                 tft.loadFont(STFONT24);
                 tft.drawString(alarmStartTime, alarmX + 60, alarmY + 20, 4);
                 tft.drawString(alarmEndTime, alarmX + 140, alarmY + 20, 4);

                 alarmY += alarmYTab;
                 alarmDisplayCount++;

                }

            } else if (alarmDateTimeInt < startStringInt) {

              strcpy(alarmDateTime, startString);
              Serial.print("next alarm at:");
              Serial.println(alarmDateTime);
              scheduleJSON.close();
              delay(20);
              return;

            }

          }


        }

      }


      // }
      scheduleJSON.close();
    }
  }
}

void drawDayLine(int numOfEntries, int x, int y, char* dayName) {

  int entryHeight = 30;
  int gapHeight = 17;

  int lineHeight = numOfEntries * entryHeight + (gapHeight * numOfEntries - 1);
  tft.drawFastVLine(x + 50, y, lineHeight, st_GRAPHITE);

  tft.setTextDatum(TR_DATUM);
  tft.setTextColor(st_LTPURPLE, st_BLACK);
  tft.loadFont(STFONT24);
  tft.drawString(dayName, x + 40, y, 4);

}
