#define FS_NO_GLOBALS
#include <FS.h>

extern "C" {
#include "crypto/base64.h"
}

#include <Wire.h>

#include <SPI.h>
#include "Adafruit_GFX.h"
#include "Adafruit_HX8357.h"
#include <TFT_eSPI.h> // Hardware-specific library for ESP8266
#include <TFT_eFEX.h>             // Include the extension graphics functions library
#include "secrets.h"
#include <WiFiClientSecure.h>
#include <PubSubClient.h>
#include <ArduinoJson.h>
#include "WiFi.h"
#include <Ticker.h>
#include "Adafruit_STMPE610.h"
#include "Adafruit_MPR121.h"
#include <Adafruit_INA219.h>
#include <Adafruit_Sensor.h>
#include "Adafruit_BME680.h"
#include <Tone32.h>

//TONE settings
#define BUZZER_PIN 13
#define BUZZER_CHANNEL 0

//BME sensor
Adafruit_BME680 bme;

//INIT INA219
Adafruit_INA219 ina219;

//INIT MPR121
#ifndef _BV
#define _BV(bit) (1 << (bit))
#endif

//Touch Sensor Init
Adafruit_MPR121 cap = Adafruit_MPR121();

// Keeps track of the last pins touched so we know when buttons are 'released'
uint16_t lasttouched = 0;
uint16_t currtouched = 0;


Ticker tickerDisplayTime;
Ticker tickerSyncTime;
Ticker tickerWeatherAni;
Ticker tickerWifiStatus;
Ticker tickerPowerReadingsDisplay;
Ticker tickerEnvReadingsDisplay;

//Pins
//MISO 19
//MOSI 18
//SCK 5
//scl 22
//SDA 23
//A35 Battery Measurement (Internal)
//TFT_CS #15
//TFT_DC #33
#define STMPE_CS 32
#define VBATPIN A13
//SD Card 14 (not used right now)
//
//Touchpins available
//pin/touch num
//4/0 ENV TOUCH (note 4 is a5)
//2/3 NOT ROUTED
//15/3 IN USE
//13/4 TONE MAYBE? (Resistor on 13 causing problems, so use for tone instead)
//12/5 MAKER TOUCH
//14/6 touch option (also for SD card, though not currently used)
//27/7 DIARY
//33/8 IN USE
//32/9 IN USE

const int debounceMS = 250;


typedef struct {
  bool laststate = false;
  bool currstate = false;;
  int x;
  int y;
  int w;
  int h;
  int val;
} tsButton;


Adafruit_STMPE610 ts = Adafruit_STMPE610(STMPE_CS); //Note, using I2C on breakboard; switch to SPI on Feather
// This is calibration data for the raw touch data to the screen coordinates
#define TS_MINX 3800
#define TS_MAXX 200  //was 100
#define TS_MINY 250 //was 100
#define TS_MAXY 3925 //3750

// Invoke TFT library
TFT_eSPI tft = TFT_eSPI();
TFT_eFEX  fex = TFT_eFEX(&tft);    // Create TFT_eFEX object "fex" with pointer to "tft" object

int tftHeight;
int tftWidth;
int staticMenuWidth = 75; //this is how much space we need for the static menu
float percContentMenuWidth = 0.85;

#define st_BLACK 0x0000
#define st_BLUE 0x9CD9 //http://www.barth-dev.de/online/rgb565-color-picker/
#define st_ORANGE 0xFCC0
#define st_LTPURPLE 0xCCD9
#define st_BLUE 0x9CD9
#define st_RED 0xCB2C
#define st_TAN 0xFE73
#define st_PURPLE 0x9CDF
#define st_PEACH 0xFCCC
#define st_HTPINK 0xCB33
#define st_WHITE 0xFFFF
#define st_OFFWHITE 0xEF7D
#define st_DKGRAY 0x8410
#define st_LTGRAY 0xA514
#define st_LTBLUE 0xC7BF
#define st_NTBLUE 0x12DD
#define st_GRAPHITE 0x31A6
#define st_YELLOW 0xFFF0
#define st_GREEN  0xAF72 //0x07F0

#define STFONT40 "StarTrek40"
#define STFONT38 "StarTrek38"
#define STFONT28 "StarTrek28"
#define STFONT24 "StarTrek24"
#define STFONT18 "StarTrek18"
#define STFONT12 "StarTrek12"


char uniqId[23];
#define PUBLISH_TOPIC   "smartDisplay/test" //"tricorder/pub"
// The MQTT topics that this device should publish/subscribe
//#define AWS_IOT_PUBLISH_TOPIC   "tricorderESP32"
//#define AWS_IOT_SUBSCRIBE_TOPIC "tricorderESP32/sub"


//// topics
const int numOfTopics    = 8;
char topics[numOfTopics][20] =
{
  "timeInfo",
  "dailyWeatherInfo",
  "hourlyWeatherInfo",
  "schedule",
  "news",
  "fitness",
  "deviceCode",
  "deviceAuth"
};

volatile boolean syncReady = false;
const char* ntpServer = "pool.ntp.org";
long  gmtOffset_sec = 0;
int   daylightOffset_sec = 0;

WiFiClientSecure httpsClient;
PubSubClient mqttClient(httpsClient);


void setDeviceId() {
  uint64_t chipid;
  chipid = ESP.getEfuseMac(); //The chip ID is essentially its MAC address(length: 6 bytes).
  uint16_t chip = (uint16_t)(chipid >> 32);
  snprintf(uniqId, 23, "TRICORDER-%04X%08X", chip, (uint32_t)chipid);
  Serial.printf(uniqId);
  //TRICORDER-BC0123A4AE30
}

boolean isAuth(int authProv) {

  char fileName[14];

  if (authProv == 0) {
    strcpy(fileName, "/msAuth");
  }

  if (authProv == 1) {
    strcpy(fileName, "/googleAuth");
  }

  File scheduleAUTH = SPIFFS.open(fileName, FILE_WRITE);

  if (!scheduleAUTH) {
    return false;
  } else {
    scheduleAUTH.close();
    return true;
  }
}

void deleteAuth(int authProv) {

  char fileName[14];

  if (authProv == 0) {
    strcpy(fileName, "/msAuth");
  }

  if (authProv == 1) {
    strcpy(fileName, "/googleAuth");
  }

  SPIFFS.remove(fileName);

}

volatile boolean menuDisplay  = false;
volatile boolean alarmActive = true;
volatile boolean beepActive = true;
volatile int brightnessVal = 100;

void buttonClickBeep() {
  if (beepActive) {
    tone(BUZZER_PIN, NOTE_F4, 100, BUZZER_CHANNEL);
    noTone(BUZZER_PIN, BUZZER_CHANNEL);
  }
}

void alarmBeep() {
  if (alarmActive) {
    tone(BUZZER_PIN, NOTE_E4, 200, BUZZER_CHANNEL);
    noTone(BUZZER_PIN, BUZZER_CHANNEL);
    tone(BUZZER_PIN, NOTE_F4, 200, BUZZER_CHANNEL);
    noTone(BUZZER_PIN, BUZZER_CHANNEL);
    tone(BUZZER_PIN, NOTE_G4, 200, BUZZER_CHANNEL);
    noTone(BUZZER_PIN, BUZZER_CHANNEL);
    tone(BUZZER_PIN, NOTE_E4, 200, BUZZER_CHANNEL);
    noTone(BUZZER_PIN, BUZZER_CHANNEL);
  }
}

void introSong() {

  float play = .9;
  float pause = .1;

  int qtr = 400 * play;
  int eighteenth = qtr / 2;
  int sixteenth = eighteenth / 2;
  int half = qtr * 2;
  int qtrD = qtr + eighteenth;
  int eighteenthD = eighteenth + sixteenth;
  int sixteenthD = sixteenth + sixteenth / 2;

  //https://github.com/robsoncouto/arduino-songs/blob/master/startrekintro/startrekintro.ino
  /*
      NOTE_D4, -8, NOTE_G4, 16, NOTE_C5, -4,
    NOTE_B4, 8, NOTE_G4, -16, NOTE_E4, -16, NOTE_A4, -16,
    NOTE_D5, 2,*/

  tone(BUZZER_PIN, NOTE_D4, eighteenth, BUZZER_CHANNEL);
  noTone(BUZZER_PIN, BUZZER_CHANNEL);
  delay(eighteenth * pause);

  tone(BUZZER_PIN, NOTE_G4, sixteenth, BUZZER_CHANNEL);
  noTone(BUZZER_PIN, BUZZER_CHANNEL);
  delay(sixteenth * pause);

  tone(BUZZER_PIN, NOTE_C5, qtrD, BUZZER_CHANNEL);
  noTone(BUZZER_PIN, BUZZER_CHANNEL);
  delay(qtrD * pause);

  tone(BUZZER_PIN, NOTE_B4, eighteenth, BUZZER_CHANNEL);
  noTone(BUZZER_PIN, BUZZER_CHANNEL);
  delay(eighteenth * pause);

  tone(BUZZER_PIN, NOTE_G4, sixteenthD, BUZZER_CHANNEL);
  noTone(BUZZER_PIN, BUZZER_CHANNEL);
  delay(sixteenthD * pause);

  tone(BUZZER_PIN, NOTE_E4, sixteenthD, BUZZER_CHANNEL);
  noTone(BUZZER_PIN, BUZZER_CHANNEL);
  delay(sixteenthD * pause);

  tone(BUZZER_PIN, NOTE_A4, sixteenthD, BUZZER_CHANNEL);
  noTone(BUZZER_PIN, BUZZER_CHANNEL);
  delay(sixteenthD * pause);

  tone(BUZZER_PIN, NOTE_D5, half, BUZZER_CHANNEL);
  noTone(BUZZER_PIN, BUZZER_CHANNEL);

}

//Indoor Temperature Readings. Save as global in order to display and check tolerances (future release)
float indoorTemp;
float indoorHum;
float indoorPres;
float indoorVOC;

void setup(void) {

  Serial.begin(115200);
  setDeviceId();

  //SPIFFS---------------
  if (!SPIFFS.begin()) {
    Serial.println("SPIFFS initialisation failed!");
    while (1) yield(); // Stay here twiddling thumbs waiting
  }
  Serial.println("\r\nSPIFFS initialised.");

  //TFT------------------------
  tft.begin();          // Initialize screen
  tft.fillScreen(st_BLACK);
  tft.setRotation(1);
  tftHeight = tft.height();
  tftWidth = tft.width();


  //Touchscreen----------------------
  if (! ts.begin(0x41)) {
    Serial.println("STMPE not found!");
    //while (1);
  }

  //MRP121----------------
  if (!cap.begin(0x5A)) {
    Serial.println("MPR121 not found, check wiring?");
    while (1);
  }
  Serial.println("MPR121 found!");
  cap.setThreshholds(14, 6);

  //INA219--------------------------------------
  ina219.begin();
  ina219.setCalibration_16V_400mA();

  //BME680--------------------------------------
  if (!bme.begin()) {
    Serial.println(F("Could not find a valid BME680 sensor, check wiring!"));
    while (1);
  }

  // Set up oversampling and filter initialization
  bme.setTemperatureOversampling(BME680_OS_8X);
  bme.setHumidityOversampling(BME680_OS_2X);
  bme.setPressureOversampling(BME680_OS_4X);
  bme.setIIRFilterSize(BME680_FILTER_SIZE_3);
  bme.setGasHeater(320, 150); // 320*C for 150 ms

  if (! bme.performReading()) {
    Serial.println("Failed to perform reading :(");
    return;
  }
  indoorTemp = (bme.temperature * 1.8) + 32;
  indoorHum = bme.humidity;
  indoorPres = bme.pressure / 100.0;
  indoorVOC = bme.gas_resistance / 1000.0;

  //Display Screens and connect to wifi-----------------------------------------
  stComm();
  lcarsStaticMenu();
  startAnimation();
  displayBATT();

  connectWifi(); //don't forget to include mqttloop in loop

  requestTimeInfo();
  requestHourlyWeather();
  requestDailyWeather();
  requestNews(NEWS_SOURCE);

  if (isAuth(0)) {
    requestSchedule();
  }

  if (isAuth(1)) {
    requestFitness();
  }

  //Play Intro Song------------------------------------------
  stopAnimation();
  displayWelcome();
  introSong();
  startAnimation();


}

void startAnimation() {
  tickerWeatherAni.attach(0.1, weatherChartAnimation);
}

void stopAnimation() {
  tickerWeatherAni.detach();
}

volatile boolean isSleep = false;
volatile boolean isDim = false;

volatile int homeInt = 0;
volatile int envInt = 0;
volatile int fitnessItem = 0;
volatile int scheduleDay = 0;

void sleepChange() {
  Serial.print("is asleep: ");
  if (!isSleep) {
    homeInt = 0;
    tft.writecommand(HX8357_SLPIN);
    isSleep = true;
    Serial.println("yes");
    digitalWrite(TFT_BL, LOW);
    displayWelcome();
  } else {
    delay(100);
    tft.writecommand(HX8357_SLPOUT);
    digitalWrite(TFT_BL, HIGH);
    isSleep = false;
    Serial.println("no");
    introSong();
  }
}

int getDutyCycle() {
  float bV = float(brightnessVal) / 100.0;
  Serial.print("duty cycle:");
  Serial.println((int)(255 * bV));
  return (int)(255 * bV);
}

//Variables for Power Measurement-----
long previousPowerMillis = 0;
long intervalPower = 1000;           // interval at which to get Power Recordings
volatile boolean powerReadingsDisplay = false;

//Variables for Environment Measurement-----
long previousIndoorMillis = 0;
long intervalIndoors = 10000;           // interval at which to get VoC Recordings

//Variables for 15 min refresh-----
long previousRefreshAMillis = 0;
long intervalRefreshA = 900000;

//Variables for 30 min refresh-----
long previousRefreshBMillis = 0;
long intervalRefreshB = 1800000;

//Variables for 5 min dim-----
long previousRefreshCMillis = 0;
long intervalRefreshC = 120000;//300000;

volatile boolean checkAlarm = false;


void loop(void) {

  mqttLoop();

  //5 min refresh Readings-------------------------------
  unsigned long currentRefreshCMillis = millis();

  if (!isSleep && (currentRefreshCMillis - previousRefreshCMillis > intervalRefreshC)) {

    Serial.println("I'm in random screens");

    int ranNum = random(7);
    previousRefreshCMillis = currentRefreshCMillis;
    stopAnimation();

    switch (ranNum) {
      case 0: //HOME
        homeInt = 0;
        displayWelcome();
        homeInt = 0;
        break;
      case 1: //SCHEMATIC
        homeInt = 1;
        displayWelcome();
        homeInt = 1;
        break;
      case 2: //WEATHER
        envInt = 1;
        displayEnviro();
        envInt = 1;
        break;
      case 3: //INDOOR
        envInt = 2;
        displayEnviro();
        envInt = 2;
        break;
      case 4: //FITNESS
        fitnessItem = 0;
        loadFitnessJSON();
        fitnessItem = 0;
        break;
      case 5: //SCHEDULE
        scheduleDay = 0;
        loadScheduleJSON(true);
        scheduleDay = 0;
        break;
      case 6: //SCHEDULE
        loadNewsJSON();
        break;
        startAnimation();

    }
  }

  currtouched = cap.touched();

  for (uint8_t i = 0; i < 12; i++) {
    // it if *is* touched and *wasnt* touched before, alert!
    if ((currtouched & _BV(i)) && !(lasttouched & _BV(i)) ) {
      Serial.print(i); Serial.println(" touched");
      previousRefreshCMillis = currentRefreshCMillis;
      stopAnimation();

      if (isSleep) {
        buttonClickBeep();
        sleepChange();
      }  else {
        switch (i) {
          case 0: //HOME
            buttonClickBeep();
            displayWelcome();
            break;
          case 1: //WEATHER
            buttonClickBeep();
            if (menuDisplay) {
              beepActive = !beepActive;
              drawKeypadVol();
              homeInt = 2;
              displayWelcome();
              break;
            }
            displayEnviro();
            break;
          case 2: //SCHEDULE
            buttonClickBeep();
            if (menuDisplay) {
              alarmActive = !alarmActive;
              drawAlarm();
              homeInt = 2;
              displayWelcome();
              break;
            }
            loadScheduleJSON(true);
            break;
          case 3: //MAKER
            buttonClickBeep();
            if (menuDisplay) {
              deleteAuth(0);
              requestDeviceCode("ms");
              break;
            }
            displayMaker();
            break;
          case 4: //NEWS
            buttonClickBeep();
            if (menuDisplay) {
              deleteAuth(1);
              requestDeviceCode("google");
              break;
            }
            loadNewsJSON();
            break;
          case 5: //FITNESS
            buttonClickBeep();
            if (menuDisplay) {
              ESP.restart();
              delay(1000);
              break;
            }
            loadFitnessJSON();
            //displayMenu(); not written yet
            break;
          case 6: //BONUS REQUEST SCHEDULE
            buttonClickBeep();
            sleepChange();
            break;
        }

      }
      startAnimation();
    }
    // if it *was* touched and now *isnt*, alert!
    if (!(currtouched & _BV(i)) && (lasttouched & _BV(i)) ) {
      Serial.print(i); Serial.println(" released");
    }
  }

  // reset our state
  lasttouched = currtouched;


  TS_Point p;

  if (ts.bufferSize()) {
    p = ts.getPoint();
  } else {
    // this is our way of tracking touch 'release'!
    p.x = p.y = p.z = -1;
  }

  if (p.z != -1) {
    stopAnimation();
    handlePress(p.x, p.y);
    startAnimation();
  }
  if (checkAlarm) {
    stopAnimation();
    loadScheduleJSON(false);
    checkAlarm = false;
    startAnimation();
  }

  //Get Indoor Readings-------------------------------
  unsigned long currentIndoorMillis = millis();

  if (currentIndoorMillis - previousIndoorMillis > intervalIndoors) {
    if (! bme.performReading()) {
      Serial.println("Failed to perform reading :(");
      return;
    }
    // save the last time you blinked the LED
    previousIndoorMillis = currentIndoorMillis;

    indoorTemp = (bme.temperature * 1.8) + 32;
    indoorHum = bme.humidity;
    indoorPres = bme.pressure / 100.0;
    indoorVOC = bme.gas_resistance / 1000.0;

  }

  //Get Power Readings---------------------------
  unsigned long currentPowerMillis = millis();
  if ( powerReadingsDisplay && (currentPowerMillis - previousPowerMillis > intervalPower)) {
    // save the last time you blinked the LED
    previousPowerMillis = currentPowerMillis;
    stopAnimation();
    displayReadings();
    startAnimation();
  }

  //15 min refresh Readings-------------------------------
  unsigned long currentRefreshAMillis = millis();

  if (currentRefreshAMillis - previousRefreshAMillis > intervalRefreshA) {

    // save the last time you blinked the LED
    previousRefreshAMillis = currentRefreshAMillis;
    requestHourlyWeather();
    displayBATT();
    if (isAuth(1)) {
      requestFitness();
    }
    if (isAuth(0)) {
      requestSchedule();
    }
    displayWIFI();
  }

  //60 min refresh Readings-------------------------------
  unsigned long currentRefreshBMillis = millis();

  if (currentRefreshBMillis - previousRefreshBMillis > intervalRefreshB) {

    // save the last time you blinked the LED
    previousRefreshBMillis = currentRefreshBMillis;
    requestDailyWeather();
    requestNews(NEWS_SOURCE);
  }

  yield();

}
